/* reader.c   by Michael Thorpe   2005-08-10 */

#include <ctype.h>
#include <stdlib.h>
#include "lisp.h"

static obj *close_paren,*dot_word,*reader_error;

static obj *macro_tick(obj *stream) {
	obj *o;

	o=read_object(incref(stream));
	if(!o) {
		if(real_eqv(exception,eoferror))
			return(throw(incref(reader_error),stream));
		decref(stream);
		return(0);
	}
	decref(stream);
	return(quoteobj(o));
}

static obj *submacro_dot(obj *stream) {
	obj *tmp1,*tmp2;

	tmp1=read_object(incref(stream));
	if(!tmp1) {
		if(real_eqv(exception,eoferror))
			return(throw(incref(reader_error),stream));
		decref(stream);
		return(0);
	}
	if(!real_eqv(tmp1,close_paren)) {
		tmp2=read_object(incref(stream));
		if(!tmp2) {
			decref(tmp1);
			if(real_eqv(exception,eoferror))
				return(throw(incref(reader_error),stream));
			decref(stream);
			return(0);
		}
		if(real_eqv(tmp2,close_paren)) {
			decref(stream);
			decref(tmp2);
			return(tmp1);
		}
		decref(tmp2);
	}
	decref(tmp1);
	return(throw(incref(reader_error),stream));
}

static obj *macro_openparen(obj *stream) {
	obj *start=0,*cur=0;
	obj *tmp;

	while(1) {
		tmp=read_object(incref(stream));
		if(!tmp) {
			if(cur) {
				CDR(cur)=incref(&nilobj);
				decref(start);
			}
			if(real_eqv(exception,eoferror))
				return(throw(incref(reader_error),stream));
			decref(stream);
			return(0);
		}
		if(real_eqv(tmp,close_paren)) {
			decref(tmp);
			break;
		}
		if(real_eqv(tmp,dot_word)) {
			decref(tmp);
			if(!cur)
				return(throw(incref(reader_error),stream));
			CDR(cur)=submacro_dot(incref(stream));
			if(!CDR(cur)) {
				CDR(cur)=incref(&nilobj);
				decref(start);
				decref(stream);
				return(0);
			}
			decref(stream);
			return(start);
		}
		if(start) {
			CDR(cur)=newobj(&cons_objtype);
			if(!CDR(cur)) {
				CDR(cur)=incref(&nilobj);
				decref(start);
				decref(stream);
				return(0);
			}
			cur=CDR(cur);
		} else {
			cur=start=newobj(&cons_objtype);
			if(!start) {
				decref(stream);
				return(0);
			}
		}
		CAR(cur)=tmp;
	}
	decref(stream);
	if(cur)
		CDR(cur)=incref(&nilobj);
	else
		start=incref(&nilobj);
	return(start);
}

static obj *macro_closeparen(obj *stream) {
	decref(stream);
	return(incref(close_paren));
}

static obj *macro_semicolon(obj *stream) {
	int c;

	while(1) {
		c=real_sgetc(stream);
		switch(c) {
		case -1:
			return(throw(incref(eoferror),stream));
		case -2:
			return(throw(incref(ioerror),stream));
		case '\n':
		case '\r':
			return(read_object(stream));
		}
	}
}

#define MAX_STRING 8192
static char string_buf[MAX_STRING+1];

static obj *macro_doublequote(obj *stream) {
	int c,i=0,seenbackslash=0;

	while(1) {
		c=real_sgetc(stream);
		if(-1==c)
			return(throw(incref(eoferror),stream));
		if(-2==c)
			return(throw(incref(ioerror),stream));
		if('\\'==c) {
			seenbackslash=!seenbackslash;
			if(seenbackslash)
				continue;
		} else if('"'==c) {
			if(!seenbackslash) {
				decref(stream);
				return(real_new_string(string_buf,i));
			}
			seenbackslash=0;
		}
		string_buf[i++]=c;
		if(MAX_STRING<i)
			return(throw(incref(reader_error),stream));
	}
}

#define MAX_TOKEN 128
static char reader_buf[MAX_TOKEN+1];

obj *read_object(obj *stream) {
	obj *o;
	int c,buflen=0,isnum=0;
	char *tmp;

	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	while(1) {
		c=real_speekc(stream);
		if(c==-1) {
			if(buflen)
				break;
			return(throw(incref(eoferror),stream));
		}
		if(!isascii(c))
			return(throw(incref(reader_error),stream));
		if(isspace(c)) {
			if(buflen)
				break;
			else {
				real_sgetc(stream);
				continue;
			}
		}
		if(buflen && (c=='"' || c=='\'' || c=='(' || c==')' || c==';'))
			break;
		real_sgetc(stream);
		switch(c) {
		case '"':
			return(macro_doublequote(stream));
		case '\'':
			return(macro_tick(stream));
		case '(':
			return(macro_openparen(stream));
		case ')':
			return(macro_closeparen(stream));
		case ';':
			return(macro_semicolon(stream));
		}
		if((!buflen || (buflen==1 && reader_buf[0]=='-')) && isdigit(c))
			isnum=1;
		if(isnum && !isdigit(c))
			isnum=0;
		reader_buf[buflen++]=c;
		if(buflen==MAX_TOKEN+1)
			return(throw(incref(reader_error),stream));
	}
	decref(stream);
	if(isnum) {
		o=newobj(&integer_objtype);
		if(!o)
			return(0);
		reader_buf[buflen]='\0';
		o->value.i=strtol(reader_buf,&tmp,10);
		if(buflen != tmp-reader_buf)
			return(throwrangecheck(o));
		return(o);
	}
	return(newword(reader_buf,buflen));
}

int reader_init() {
	if(!(close_paren=newword(")",1))
	|| !(dot_word=newword(".",1))
	|| store_builtin_ns1("read-object",read_object)
	|| !(reader_error=newword("reader-error",12)))
		return(1);
	INC_KNOWN_OBJECTS(3);
	return(0);
}

