/* string.c   by Michael Thorpe   2005-08-10 */

#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include "lisp.h"

static void drop_string(obj *o) {
	free(o->value.s.data);
}

static obj *get_char(obj *str,obj *idx) {
	obj *c;

	if(str->type != &string_objtype && str->type != &word_objtype) {
		decref(idx);
		return(throwtypeerror(str));
	}
	if(idx->type != &integer_objtype) {
		decref(str);
		return(throwtypeerror(idx));
	}
	if(idx->value.i<0 || str->value.s.len<=idx->value.i) {
		decref(str);
		return(throwrangecheck(idx));
	}
	c=newobj(&integer_objtype);
	if(c)
		c->value.i=(unsigned char)str->value.s.data[idx->value.i];
	decref(str);
	decref(idx);
	return(c);
}

static obj *put_char(obj *str,obj *idx,obj *c) {
	if(str->type != &string_objtype) {
		decref(idx);
		decref(c);
		return(throwtypeerror(str));
	}
	if(idx->type != &integer_objtype) {
		decref(str);
		decref(c);
		return(throwtypeerror(idx));
	}
	if(idx->value.i<0 || str->value.s.len<=idx->value.i) {
		decref(str);
		decref(c);
		return(throwrangecheck(idx));
	}
	if(c->type != &integer_objtype) {
		decref(str);
		decref(idx);
		return(throwtypeerror(c));
	}
	if(c->value.i<0 || UCHAR_MAX<c->value.i) {
		decref(str);
		decref(idx);
		return(throwrangecheck(c));
	}
	str->value.s.data[idx->value.i]=c->value.i;
	decref(idx);
	decref(c);
	return(str);
}

static obj *string_to_integer(obj *str) {
	obj *i;
	char buf[32];
	char *s;
	unsigned long l;

	if(str->type != &string_objtype || str->value.s.len>sizeof(buf)-1)
		return(throwtypeerror(str));
	memcpy(buf,str->value.s.data,str->value.s.len);
	buf[str->value.s.len]='\0';
	l=strtol(buf,&s,0);
	if(s != buf+str->value.s.len)
		return(throwtypeerror(str));
	if(l==LONG_MIN || l==LONG_MAX)
		return(throwrangecheck(str));
	decref(str);
	i=newobj(&integer_objtype);
	if(i)
		i->value.i=l;
	return(i);
}

static obj *string_to_symbol(obj *str) {
	obj *sym;

	if(str->type != &string_objtype)
		return(throwtypeerror(str));
	sym=newword(str->value.s.data,str->value.s.len);
	decref(str);
	return(sym);
}

static obj *string_tostring(obj *in) {
	obj *o;
	int i,j;

	for(i=j=0;i<in->value.s.len;i++) {
		switch(in->value.s.data[i]) {
		case '\0': case '\\': case '"': j++; break;
		}
	}
	o=real_new_string(0,in->value.s.len+2+j);
	if(!o)
		return(throwoom());
	o->value.s.data[0]='"';
	for(i=0,j=1;i<in->value.s.len;i++) {
		switch(in->value.s.data[i]) {
		case '\0':
			o->value.s.data[j++]='\\';
			o->value.s.data[j++]='0';
			break;
		case '\\':
		case '"':
			o->value.s.data[j++]='\\';
		default:
			o->value.s.data[j++]=in->value.s.data[i];
			break;
		}
	}
	o->value.s.data[j++]='"';
	assert(o->value.s.len==j);
	decref(in);
	return(o);
}

static obj *string_length(obj *str) {
	obj *l;

	if(str->type != &string_objtype)
		return(throwtypeerror(str));
	l=newobj(&integer_objtype);
	if(l)
		l->value.i=str->value.s.len;
	decref(str);
	return(l);
}

static obj *substring(obj *str,obj *offobj,obj *lenobj) {
	obj *substr;
	unsigned long off,len;

	if(str->type != &string_objtype) {
		decref(offobj);
		decref(lenobj);
		return(throwtypeerror(str));
	}
	if(offobj->type != &integer_objtype) {
		decref(str);
		decref(lenobj);
		return(throwtypeerror(offobj));
	}
	if(lenobj->type != &integer_objtype) {
		decref(str);
		decref(offobj);
		return(throwtypeerror(lenobj));
	}
	off=offobj->value.i;
	len=lenobj->value.i;
	if(off<0)
		off+=str->value.s.len;
	if(len<=0)
		len=str->value.s.len-off-len;
	if(off<0 || off>str->value.s.len) {
		decref(lenobj);
		decref(str);
		return(throwrangecheck(offobj));
	}
	decref(offobj);
	if(len<0) {
		decref(str);
		return(throwrangecheck(lenobj));
	}
	decref(lenobj);
	if(off+len<0 || off+len>str->value.s.len) {
		decref(str);
		return(throwrangecheck(incref(&nilobj)));
	}
	substr=real_new_string(str->value.s.data+off,len);
	decref(str);
	return(substr);
}

obj *real_new_string(const char *data,unsigned long len) {
	obj *o;
	char *s;

	s=(char *)malloc(len);
	if(!s)
		return(throwoom());
	o=newobj(&string_objtype);
	if(!o) {
		free(s);
		return(0);
	}
	o->value.s.data=s;
	o->value.s.len=len;
	if(data)
		memcpy(o->value.s.data,data,len);
	else
		memset(o->value.s.data,0,len);
	return(o);
}

static obj *new_string(obj *o) {
	obj *r;

	if(o->type != &integer_objtype)
		return(throwtypeerror(o));
	if(o->value.i<0)
		return(throwrangecheck(o));
	r=real_new_string(0,o->value.i);
	decref(o);
	return(r);
}

int string_init() {
	if(store_builtin_ns2("get-char",get_char)
	|| store_builtin_ns1("new-string",new_string)
	|| store_builtin_ns3("put-char",put_char)
	|| store_builtin_ns1("string->integer",string_to_integer)
	|| store_builtin_ns1("string->symbol",string_to_symbol)
	|| store_builtin_ns1("string-length",string_length)
	|| store_builtin_ns3("substring",substring))
		return(1);
	return(0);
}

objtype string_objtype={
	name: "string",
	tostring: string_tostring,
	dropobj: drop_string
};

